import java.io.* ;
/*
 AJGrep

 Luc Maranget, projet Moscova, Inria Rocquencourt
 
 Copyright 2006 Institut National de Recherche en Informatique et    
 en Automatique.  All rights reserved.  This file is distributed     
 under the terms of the Q Public License version 1.0.          
*/

class Auto {
  private final static int initSize = 32 ;

  int fstate ;
  Transition [] trans ;
  boolean [] sharp ;
  boolean [] dot ;
  Epsilon [] epsilon ;

  boolean someSharp () {
    for (int i=0 ; i < fstate ; i++)
      if (sharp[i]) return true ;
    return false ;
  }

  boolean someEpsilon() {
    for (int i=0 ; i < fstate; i++)
      if (epsilon[i] != null) return true ;
    return false ;
  }

  // Analyse syntaxique
  private Auto() {
    fstate = 0 ;
    trans = new Transition [initSize] ;
    sharp = new boolean [initSize] ;
    dot = new boolean [initSize] ;
    trans = new Transition [initSize] ;
    epsilon = new Epsilon [initSize] ;
  }

  private static void syntax(String pat, int i, String msg) {
    System.err.println("Syntax error: " + msg) ;
    System.err.println(pat + ", pos=" + i) ;
    System.exit(2) ;
  }

  private int simple(String pat, int i) {
    if (i >= pat.length()) return i ;
    char c = pat.charAt(i) ;

    if (c == '\\') {
      i++ ;
      if (i >= pat.length()) syntax(pat, i, "``\\'' at end") ;
      c = pat.charAt(i) ;
    }
 
    if (c == '(') {
      i = or(pat, i+1) ;
      if  (i >= pat.length() || pat.charAt(i) != ')')
        syntax(pat, i, "``)'' is missing") ;
      return i+1 ;
    } else {
      switch (c) {
      case '#':
        sharp() ;
        return i+1 ;
            
      case '.':
        dot() ;
        return i+1 ;

      case ')': case '|': case '(' :
      case '?': case '*': case '+' :
        return i ;
        
      default:
        append(c) ;
        return i+1 ;
      }
    }
  }


  private int postfixe(String pat, int i) {
    i = simple(pat, i) ;
    while (i < pat.length()) {
      char c = pat.charAt(i) ;
      if (c == '*') {
        star() ; i++ ;
      } else if (c == '+') {
        plus() ; i++ ;
      } else if (c == '?') {
        optional() ; i++ ;
      } else
        return i ;
    }
    return i ;
  }

  private int concat(String pat, int i) {
    do {
      Auto a = new Auto() ;      
      i = a.postfixe(pat, i) ;
      append(a) ;
    } while
        (i < pat.length() && pat.charAt(i) != ')' && pat.charAt(i) != '|') ;
    return i ;
  }

  private int or(String pat, int i) {
    i = concat(pat, i) ;
    if (i < pat.length() && pat.charAt(i) == '|') {
      Auto a = new Auto () ;
      i = a.or(pat,i+1) ;
      alt(a) ;
    }
    return i ;
  }

  Auto (String pat) {
    fstate = 0 ;
    trans = new Transition [initSize] ;
    sharp = new boolean [initSize] ;
    dot = new boolean [initSize] ;
    trans = new Transition [initSize] ;
    epsilon = new Epsilon [initSize] ;
    int i = or(pat, 0) ;
    if (i < pat.length())
      syntax(pat, i, "End expected") ;
  }

  private void realloc(int targetSize) {
    int t ;

    for (t = trans.length ; t < targetSize ; t *= 2)
      ;
    if (t == trans.length) return ;

    Transition [] newTrans = new Transition [t] ;
    boolean [] newSharp = new boolean [t] ;
    boolean [] newDot = new boolean[t] ;
    Epsilon [] newEpsilon = new Epsilon [t] ;
    for (int i = 0 ; i <= fstate ; i++) {
      newTrans[i] = trans[i] ;
      newSharp[i] = sharp[i] ;
      newDot[i] = dot[i] ;
      newEpsilon[i] = epsilon[i] ;      
    }
    trans = newTrans ;
    sharp = newSharp ;
    dot = newDot ;
    epsilon = newEpsilon ;
  }
    
  private void addState() {
    realloc(fstate+1) ;
    fstate++ ;
  }

  private void addEpsilon(int from, int target) {
    epsilon[from] = new Epsilon(target, epsilon[from]) ;
  }

  private void addTransition(int from, char c) {
    trans[from] = new Transition(c, trans[from]) ;
  }

  void append(char c) {
    addState() ;
    addTransition(fstate-1, c) ;
  }

  // There is a transition from final state
  private boolean fromFinal () {
    return epsilon[fstate] != null || sharp[fstate] ;
  }

  // There is a transition to start state
  private boolean toStart() {
    if (sharp[0]) return true ;
    for (int i = 0 ; i <= fstate ; i++) {
      for (Epsilon p = epsilon[i] ; p != null ; p = p.next)
        if (p.target == 0) return true ;
    }
    return false ;
  }

  void append(Auto a) {
    // Ajouter une transition vide (et sa cible) si ncessaire
    if (fromFinal() && a.toStart()) {
      if (sharp[fstate] && a.sharp[0])
        return ;
      addState() ;
      addEpsilon(fstate-1, fstate) ;
    }
    realloc(fstate+a.fstate+2) ;

    // Fusionner l'tat final avec l'tat initial de a
    trans[fstate] = a.trans[0] ;
    sharp[fstate] |= a.sharp[0] ;
    dot[fstate] |= a.dot[0] ;
    for (Epsilon p = a.epsilon[0] ; p != null ; p = p.next)
        addEpsilon(fstate, p.target+fstate) ;

    // Recopier ensuite a en dcalant tout de fstate
    for (int i = 1 ; i <= a.fstate ; i++) {
      trans[i+fstate] = a.trans[i] ;
      sharp[i+fstate] = a.sharp[i] ;
      dot[i+fstate] = a.dot[i] ;
      epsilon[i+fstate] = null ; // Sait-on jamais
      for (Epsilon p = a.epsilon[i] ; p != null ; p = p.next)
        addEpsilon(i+fstate, p.target+fstate) ;
    }
    fstate += a.fstate ;
  }

  void sharp() {
    sharp[fstate] = true ;
  }

  void dot() {
    addState() ;
    dot[fstate-1] = true ;
  }

  void star() {
    addEpsilon(0,fstate) ;
    addEpsilon(fstate,0) ;
  }

  void optional() {
    addEpsilon(0,fstate) ;
  }

  void plus() {
    addEpsilon(fstate,0) ;
  }


  void alt(Auto a) {
    int f1 = fstate+1 ;
    int e2 = f1+1 ;
    int f2 = e2+a.fstate ;
    int f = f2+1 ;
    realloc(f+1) ;
    
    // Translater l'automate de un cran  droite
    for (int i = fstate ; i >=0 ; i--) {      
      trans[i+1] = trans[i] ;
      sharp[i+1] = sharp[i] ;
      dot[i+1] = dot[i] ;
      epsilon[i+1] = null ;
      for (Epsilon p = epsilon[i] ; p != null ; p = p.next)
        addEpsilon(i+1, p.target+1) ;
    }

    // Ajouter les transitions concernant cette premire alternative
    trans[0] = null ;
    epsilon[0] = null ;
    dot[0] = false ;
    sharp[0] = false ;
    addEpsilon(0,1) ;
    addEpsilon(f1,f) ;

    // Recopier a  partir de la position e2
    for (int i=0 ; i <= a.fstate ; i++) {
      trans[i+e2] = a.trans[i] ;
      sharp[i+e2] = a.sharp[i] ;
      dot[i+e2] = a.dot[i] ;
      epsilon[i+e2] = null ;
      for (Epsilon p = a.epsilon[i] ; p != null ; p = p.next)
        addEpsilon(i+e2, p.target+e2) ;
    }

    // Ajouter les transitions concernant cette seconde alternative
    addEpsilon(0,e2) ;
    addEpsilon(f2,f) ;
    fstate = f ;
  }


  // Les ``#'' sont inconnus des automates de niveau C, les transformer en
  // ``.*''
  // Un peu lourd, vaut mieux se passer du niveau B au fond
  void sharpOut() {
    for (int i=0 ; i <= fstate ; i++) {
      if (sharp[i]) {
        sharp[i] = false ;
        addState() ;
        for (int j = fstate ; j > i ; j--) {
          trans[j] = trans[j-1] ;
          sharp[j] = sharp[j-1] ;
          epsilon[j] = null ;
          for (Epsilon p = epsilon[j-1] ; p != null ; p = p.next) {
            addEpsilon(j,p.target >= i ? p.target+1 : p.target) ;
          }
        }
        for (int j = i-1 ; j >= 0 ; j--) {
           for (Epsilon p = epsilon[j] ; p != null ; p = p.next)
             p.target = p.target > i ? p.target+1 : p.target ;
        }
        trans[i] = null ;
        epsilon[i] = null ;
        dot[i] = true ;
        addEpsilon(i,i+1) ;
        addEpsilon(i+1,i) ;
      }
    }
  }

  public String toString() {
    StringBuffer r = new StringBuffer () ;

    r.append ("fstate = " + fstate + "\n") ;
    for (int i = 0 ; i <= fstate ; i++) {
      r.append(i + ": " + trans[i] + "; " + epsilon[i]) ;
      if (dot[i]) r.append ("; DOT ") ;
      if (sharp[i]) r.append ("; SHARP ") ;
      r.append("\n") ;
    }
    return r.toString() ;
  }
}
